import re
from functools import partial
from typing import Protocol, List


class Splitter(Protocol):
    def __call__(self, text: str, maxsplit: int = 0) -> List[str]:
        raise NotImplementedError()


def regex_splitter(text: str, split_pattern=r'\s+', strip_before_split=True, maxsplit=0):
    return re.split(
        split_pattern,
        text.strip() if strip_before_split else text,
        maxsplit=maxsplit
    )


def create_regex_splitter(split_pattern=r'\s+', strip_before_split=True) -> Splitter:
    # noinspection PyTypeChecker
    return partial(
        regex_splitter,
        split_pattern=split_pattern,
        strip_before_split=strip_before_split
    )


default_splitter = create_regex_splitter(
    split_pattern=r'\s+',
    strip_before_split=True
)



import json
import os
import sys

from pyshrimp import run
from pyshrimp._internal.wrapper.magicwrapper_state import _MAGIC_STATE


def _split_items_by(items, separator_element):
    left = []
    right = []
    separator_found = False
    for el in items:
        if separator_found:
            right.append(el)

        elif el == separator_element:
            separator_found = True

        else:
            left.append(el)

    return left, right


def main():
    opts = json.loads(os.environ.get('PYSHRIMP_MAGIC_WRAPPER_OPTS', '{}'))

    original_args = sys.argv
    wrapper_args, program_args = _split_items_by(original_args[1:], '--')
    program_file = program_args[0]
    sys.argv = program_args.copy()

    _MAGIC_STATE.active = True

    def _re_run_script():
        args = [sys.executable] + original_args
        os.execlp(args[0], *args)

    # initialize
    def _script():
        # load script
        import importlib.util
        module_spec = importlib.util.spec_from_file_location("__main__", program_file)
        module_ref = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module_ref)

    run(
        script=_script,
        check_main=False,
        devloop=opts.get('devloop') == 'true',
        elevate=opts.get('elevate') == 'true',
        re_run_script=_re_run_script,
        script_path=program_file
    )


if __name__ == '__main__':
    main()



class IllegalStateException(Exception):

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class IllegalArgumentException(Exception):

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class AsyncFunctionInvocationException(BaseException):

    def __init__(self, *args: object) -> None:
        super().__init__(*args)




import os
import threading


class FileBasedLockInternal:

    def __init__(self, lock_file_path):
        self._lock_file_path = lock_file_path

    def _get_lock_info(self) -> str:
        return f'PID: {os.getpid()} THREAD: {threading.current_thread().ident}'

    def try_lock(self):
        raise NotImplementedError()

    def release_lock(self):
        raise NotImplementedError()





import datetime
import hashlib
import inspect
import logging
import os
import subprocess
import sys
import time
import traceback
from builtins import KeyboardInterrupt

from time import sleep
from typing import Callable

# noinspection PyProtectedMember
from pyshrimp._internal.utils.errors import _exit_error
from pyshrimp._internal.wrapper.magicwrapper_state import _is_magic_active


def _init_logging():
    # TODO: consider level coloring...
    logging.basicConfig(
        format='%(asctime)s.%(msecs)03d %(levelname)8s | %(message)s',
        level=logging.INFO,
        stream=sys.stdout,
        datefmt="%Y-%m-%d %H:%M:%S"
    )


def _re_run_script():
    args = [sys.executable] + sys.argv
    os.execlp(args[0], *args)


def _calculate_file_checksum(script_path):
    with open(script_path, 'rb') as f:
        return hashlib.sha1(f.read()).hexdigest()


def _elevate_with_sudo(devloop: bool):
    sudo_path = os.environ.get('PYSHRIMP_SUDO_PATH', 'sudo')
    os.environ['PYTHONPATH'] = ':'.join(sys.path)
    args = [sudo_path, '-E', '--preserve-env=PATH,PYTHONPATH', '--', sys.executable] + sys.argv

    if devloop:
        # cannot use exec, run sub-process
        return subprocess.run(args).returncode
    else:
        os.execlp(args[0], *args)
        # this will be never invoked as we exec:
        return -1


def _run(
    script: Callable,
    elevate=False,
    init_logging=_init_logging,
    devloop=False,
    check_main=True,
    re_run_script=_re_run_script,
    script_path=None
):
    if check_main and inspect.currentframe().f_back.f_locals['__name__'] != '__main__':
        return

    if _is_magic_active() and not inspect.currentframe().f_back.f_code.co_filename.endswith('magicwrapper.py'):
        raise _exit_error(
            'Looks like "run(...)" was invoked manually while running with magic enabled. '
            'This will likely cause unexpected behavior and hence such invocation is not supported. '
            'Please either remove explicit "run(...)" invocation or the "magic" from $opts.'
        )

    # remember file checksum before executing
    main_script_path = os.path.abspath(script_path or sys.modules['__main__'].__file__)
    main_script_checksum = _calculate_file_checksum(main_script_path) if devloop else None
    main_script_change_time = os.stat(main_script_path).st_mtime
    execution_start = time.time()

    if elevate and os.getuid() != 0 and os.environ.get('PYSHRIMP_RUNNING_UNDER_FAKE_SUDO', 'false') != 'true':
        exit_code = _elevate_with_sudo(devloop)

    else:
        if init_logging:
            init_logging()

        # noinspection PyBroadException
        try:
            exit_code = script() or 0
        except KeyboardInterrupt:
            exit_code = 130
        except SystemExit as ex:
            exit_code = ex.code
        except BaseException:
            if devloop:
                print('ERROR: Got exception while running script:')
                traceback.print_exc()
                exit_code = 1
            else:
                raise

    if devloop:
        # TODO: wait until script changes or user input is given
        print()
        print(f' /' + '-' * 60)
        print(f' | Script finished, exit code: {exit_code}')
        print(f' | Execution took: {datetime.timedelta(seconds=time.time() - execution_start)}')
        print(f' | Waiting until file changes:')
        print(f' |   {main_script_path}')
        sys.stdout.flush()

        try:
            last_mtime = main_script_change_time
            while True:
                current_mtime = os.stat(main_script_path).st_mtime
                if last_mtime != current_mtime:
                    last_mtime = current_mtime
                    if main_script_checksum != _calculate_file_checksum(main_script_path):
                        break

                sleep(0.2)
        except KeyboardInterrupt:
            sys.exit(130)

        print(f' | File changed, re-running script * at: {datetime.datetime.now()}')
        print(f' /' + '-' * 60)
        print()
        sys.stdout.flush()
        re_run_script()
    else:
        sys.exit(exit_code)



from multiprocessing.pool import AsyncResult, ThreadPool

__thread_pools = {}


def __get_thread_pool(name) -> ThreadPool:
    if name not in __thread_pools:
        __thread_pools[name] = ThreadPool(10)

    return __thread_pools[name]


def in_background(fn) -> AsyncResult:
    return __get_thread_pool('default').apply_async(fn)



import os
from contextlib import contextmanager
from typing import Type

from pyshrimp._internal.locking.file_based_lock_internal import FileBasedLockInternal


def get_file_based_lock_impl_for_posix() -> Type[FileBasedLockInternal]:
    import fcntl
    class FileBasedLockPosix(FileBasedLockInternal):

        def __init__(self, lock_file_path):
            self._lock_file_path = lock_file_path
            self._lock_fd = None

        def try_lock(self):
            if self._lock_fd is not None:
                raise Exception('Cannot lock twice!')

            fd = os.open(
                self._lock_file_path,
                os.O_RDWR | os.O_CREAT | os.O_APPEND
            )

            try:
                fcntl.flock(
                    fd,
                    # obtain exclusive lock, fail if unable to lock
                    fcntl.LOCK_EX | fcntl.LOCK_NB
                )
            except (IOError, OSError):
                # failed to lock
                os.close(fd)
                return False

            with open(self._lock_file_path, 'w') as f:
                f.write(self._get_lock_info())

            # lock acquired
            self._lock_fd = fd

            return True

        def release_lock(self):
            self._assert_locked()

            # clear lock file
            open(self._lock_file_path, 'w').close()

            # release lock
            os.unlink(self._lock_file_path)
            fd = self._lock_fd
            self._lock_fd = None
            fcntl.flock(fd, fcntl.F_UNLCK)
            os.close(fd)

        def _assert_locked(self):
            if self._lock_fd is None:
                raise Exception('Illegal state: not locked')

        @contextmanager
        def acquire_lock(self):
            locked = self.try_lock()
            try:
                yield locked
            finally:
                if locked:
                    self.release_lock()

    return FileBasedLockPosix